/************************************************************
*
* REXLANG - Using the Sandbox MH-Z16 NDIR CO2 Sensor via I2C
* Written by Eric Novikoff
* Version 1.0 - 2020.04.21
* 
* Design Overview:
*    Reads the MH-Z16 sensor and returns values via the SC16IS750IPW I2C-to-serial chip.
* Usage Notes:
*    It is recommended to run this driver in a separate task that operates at a longer
*    tick length than your project to avoid stretching a short tick time with the included
*    sleep statements.  This tick time MUST also be longer than the RECEIVE_TIMEOUT for 
*    proper operation (though this could be coded around in this driver if necessary.)
* 
*    To calibrate the sensor, apply a rising edge to the Input(0) pin.  Only calibrate when
*    the sensor is exposed to fresh air, ideally outside and not installed in its operating
*    location (unless that is outside too!  Calibrating when the CO2 concentration is not 
*    outdoor ambient will give VERY wrong readings!
*
*************************************************************/

/*
 * Device Constants
 */
#define MHZ16_I2C_ADDRESS 	0x4D

//SC16IS750 General Registers
#define  RHR        		0x00
#define  THR      		  	0x00
#define  IER     		 	0x01
#define  FCR        		0x02
#define  IIR        		0x02
#define  LCR        		0x03
#define  MCR        		0x04
#define  LSR        		0x05
#define  MSR        		0x06
#define  SPR        		0x07
#define  TCR        		0x06
#define  TLR        		0x07
#define  TXLVL      		0x08
#define  RXLVL      		0x09
#define  IODIR      		0x0A
#define  IOSTATE    		0x0B
#define  IOINTENA   		0x0C
#define  IOCONTROL  		0x0E
#define  EFCR       		0x0F

//SC16IS750 Special Registers
#define  DLL        		0x00
#define  DLH        		0x01

//SC16IS750 Enhanced Registers
#define  EFR        		0x02
#define  XON1       		0x04
#define  XON2       		0x05
#define  XOFF1      		0x06
#define  XOFF2      		0x07

//SC16IS750 Bitmasks
#define  SRESET					0x08
#define  FCR_FIFO_RESET_ENABLE	0x07
#define  LCR_DIV_8_N_1			0x83
#define	 LCR_8_N_1				0x03 //8-N-1 data
#define  BAUD_CLOCK_DIV			0x60 //9600 baud

//Application Related
#define  SC16IS750_CRYSTAL_FREQ 14745600  // 14.745600 MHz  Useful but unused info
// Driver waits for input from MH-Z16 for the amount of seconds in this timeout constant
// *** Make sure your tick length is longer than this timeout or you will get an error -1
// *** indicating that the driver could not complete execution in the tick time.
#define  RECEIVE_TIMEOUT	1 // Seconds to try to get data from MH-Z16

//Rexlang Related
#define  REXOK			0
#define  REXERR			1

#define  TRUE				0xFF
#define  FALSE				0x00

/*
 * REXLANG Parameters and inputs/outputs
 */
string parameter(0) i2c_dev; // the I2C bus is defined by the p0 parameter
long parameter(1) i2c_addr; // Overrides the default setting above if nonzero.  Note this is a DECIMAL parameter!

//assigning variables to outputs
double output(0) co2_ppm; //value to send to y0 output
long output(1) cal_running;

//assigning inputs to variables
long input(0) calibrate; //trigger for calibration, edge sensitive

//declaration of global variables
long i2c_bufTx[10]; //buffer for transmitting data
long i2c_bufRx[10]; //buffer for receiving data
long i2c_bus_handle;
long i2c_chip_address;
long i2c_write_count;
long i2c_read_count;
long i2c_ret_fun;
long buf[9];
double ppm;
long calInProgress = FALSE;

//MH-Z16 commands
long cmd_measure[9];
long cmd_calibrateZero[9];
long cmd_enableAutoCalibration[9];
long cmd_disableAutoCalibration[9];


long read_register(long addr) // SC16IS750 register read
{
    Trace(1, "readreg");
    Trace(1, addr);
    
    i2c_bufTx[0] = addr << 3; //register no.
    i2c_write_count = 1;
    i2c_read_count = 1;
    i2c_ret_fun = I2C(i2c_bus_handle, i2c_chip_address, i2c_bufTx, i2c_write_count, i2c_bufRx, i2c_read_count);
    Trace(1, i2c_bufRx[0]);
    return i2c_bufRx[0] & 0xff;
}

long write_register(long addr, long data) // SC16IS750 register write
{
    Trace(2, "writereg");
    Trace(2, addr);
    Trace(2, data);
    
    i2c_bufTx[0] = addr << 3; //register no. 
    i2c_bufTx[1] = data & 0xFF;
    i2c_write_count = 2;
    i2c_read_count = 0;
    i2c_ret_fun = I2C(i2c_bus_handle, i2c_chip_address, i2c_bufTx, i2c_write_count, i2c_bufRx, i2c_read_count);
    Trace(201,"i2c_ret_fun="+long2str(i2c_ret_fun,16));
    return i2c_ret_fun;
}

long send_mhz16(long cmdbuf[], long n)
{
    long retries = 10;
    
    Trace(3, "sendmz16");

    while (retries) {
        if (read_register(TXLVL) >= n) {
            i2c_bufTx[0] = THR << 3; // initiate write to FIFO
            copyarray(9,cmdbuf,0,i2c_bufTx,1);  // send rest of bytes, the command string
            i2c_write_count = n+1;
            i2c_read_count = 0;
            Trace(301,"write i2c@"+long2str(i2c_bufTx[0],16)+"#"+long2str(n,16)+":"+long2str(i2c_bufTx[1],16)+","+long2str(i2c_bufTx[2],16)+","+long2str(i2c_bufTx[3],16)+","+long2str(i2c_bufTx[4],16));
            i2c_ret_fun = I2C(i2c_bus_handle, i2c_chip_address, i2c_bufTx, i2c_write_count, i2c_bufRx, i2c_read_count);
            break;
        } else retries--;
    }
    Trace(302,retries);
    if (retries) return FALSE; else return TRUE;
}

long receive_mhz16(long n) {
    
    long  i, pbufi = 0;
    long  rx_level;
    double start = CurrentTime();
    
    Trace(4, "recieve_mhz16");

    while (n) {
        rx_level = read_register(RXLVL);
        if (rx_level > n) rx_level = n;
        Trace(401,"rxlevel="+long2str(rx_level));
        if (rx_level) {
            i2c_bufTx[0] = RHR << 3; // initiate read from FIFO
            i2c_write_count = 1;
            Trace(402,"write i2c:"+long2str(i2c_bufTx[0],16));
            i2c_read_count = rx_level;
            i2c_ret_fun = I2C(i2c_bus_handle, i2c_chip_address, i2c_bufTx, i2c_write_count, i2c_bufRx, i2c_read_count);
            Trace(403,"read i2c #"+long2str(i2c_read_count,16)+": "+long2str(i2c_bufRx[0],16)+","+long2str(i2c_bufRx[1],16)+","+long2str(i2c_bufRx[2],16)+","+long2str(i2c_bufRx[3],16));
            for (i=0; i<rx_level; i++) {
                buf[pbufi++] = i2c_bufRx[i];
                n--;
            }
        }
        Trace(404,(ElapsedTime(CurrentTime(),start)));
        if (ElapsedTime(CurrentTime(), start) > RECEIVE_TIMEOUT) {
            Trace(402,"timeout!");
            return TRUE;
        }
        Sleep(0.1); //slow down the retries
    }
    return FALSE;
}

long parse_mhz16(void)
{
    
    long i;
    long checksum = 0;
    
    Trace(5, "parse");

    for (i=0; i<9; i++) {
        checksum += buf[i];
    }
    Trace(501,"checksum="+long2str(checksum,10));
    Trace(502,"buf="+long2str(buf[0],16)+","+long2str(buf[1],16));

    if ((buf[0] == 0xFF) && (buf[1] == 0x9C) && ((checksum & 0xFF) == 0xFF)) {
        ppm = (double)(buf[2] << 24) + (double)(buf[3] << 16) + (double)(buf[4] << 8) + (double)buf[5];
        Trace(503,"ppm="+double2str(ppm));
        return FALSE;
    } else {
        return TRUE;
    }
}

//the init procedure is executed once when the REXLANG function block initializes
long init(void)
{
    long retries = 10;
    
    Trace(0,"init");
    
    // set up MH-Z16 command strings
    //cmd_measure                     = {0xFF,0x01,0x9C,0x00,0x00,0x00,0x00,0x00,0x63};
    cmd_measure[0]=0xFF;cmd_measure[1]=0x01;cmd_measure[2]=0x9C;cmd_measure[3]=0x00;cmd_measure[4]=0x00;
    cmd_measure[5]=0x00;cmd_measure[6]=0x00;cmd_measure[7]=0x00;cmd_measure[8]=0x63;
    //cmd_calibrateZero               = {0xFF,0x01,0x87,0x00,0x00,0x00,0x00,0x00,0x78};
    cmd_calibrateZero[0]=0xFF;cmd_calibrateZero[1]=0x01;cmd_calibrateZero[2]=0x87;cmd_calibrateZero[3]=0x00;
    cmd_calibrateZero[4]=0x00;cmd_calibrateZero[5]=0x00;cmd_calibrateZero[6]=0x00;cmd_calibrateZero[7]=0x00;
    cmd_calibrateZero[8]=0x78;
    //cmd_enableAutoCalibration  = {0xFF,0x01,0x79,0xA0,0x00,0x00,0x00,0x00,0xE6};
    //cmd_disableAutoCalibration = {0xFF,0x01,0x79,0x00,0x00,0x00,0x00,0x00,0x86};
    cmd_disableAutoCalibration[0]=0xFF;cmd_disableAutoCalibration[1]=0x01;cmd_disableAutoCalibration[2]=0x79;
    cmd_disableAutoCalibration[3]=0x00;cmd_disableAutoCalibration[4]=0x00;cmd_disableAutoCalibration[5]=0x00;
    cmd_disableAutoCalibration[6]=0x00;cmd_disableAutoCalibration[7]=0x00;cmd_disableAutoCalibration[8]=0x86;

    
	// Setup I2C
    i2c_bus_handle = OpenI2C(i2c_dev); // open I2C bus
    if (i2c_addr) i2c_chip_address = i2c_addr; // 7-bit address of the I2C device
    else i2c_chip_address = MHZ16_I2C_ADDRESS;
    
    // Reset SC16IS740
    write_register(IOCONTROL, SRESET);
    
    // Initialize SC16IS740
    while (retries) {
        if (!write_register(FCR, FCR_FIFO_RESET_ENABLE)) {
            if (!write_register(LCR, LCR_DIV_8_N_1)) {
                if (!write_register(DLL, BAUD_CLOCK_DIV)) {
                    if (!write_register(DLH, BAUD_CLOCK_DIV >> 8)) {
                        if (!write_register(LCR, LCR_8_N_1)) {
                            break;
                        }
                    }
                }
            }
        }
        retries--;
    }
    
    if (!retries) return REXERR;
    
    // Turn off autocalibration
    if (!write_register(FCR, FCR_FIFO_RESET_ENABLE)) {
            if (!send_mhz16(cmd_disableAutoCalibration, 9)) {
                Sleep(0.1);
                return REXOK;
            }
        }
    
    return REXERR;
}

//the main procedure is executed repeatedly (once in each sampling period)
long main(void)
{
    double tmp;
    
    Trace(1,"main");
    
    if (!calibrate) {
        calInProgress = FALSE;
        cal_running = FALSE;
        write_register(FCR, FCR_FIFO_RESET_ENABLE);
        //tmp = sqrt(2.0); //waste a few microseconds
        if (!send_mhz16(cmd_measure, 9)) {
            Sleep(0.001);
            if (!receive_mhz16(9)) {
                if (!parse_mhz16()) {
                    co2_ppm = ppm;
                    return REXOK;
                }
            }
        }
    } 
    else {
    //calibrate sensor.  Note: sensor must be outside, not in a room or greenhouse!
        if (!calInProgress)
        {
            calInProgress = TRUE;
            cal_running = TRUE;
            if (!write_register(FCR, FCR_FIFO_RESET_ENABLE)) {
                if (!send_mhz16(cmd_calibrateZero, 9)) {
                    Sleep(0.2);
                    return REXOK;
                }
            }
        }
    }
	return REXERR;
}

//the exit procedure is executed once when the task is correctly terminated
// (system shutdown, downloading new control algorithm, etc.)
long exit(void)
{
	return 0;
}
